/*!
 * \file
 * \brief   Distributed bitonic implementation.
 *
 * \author
 *    Christos Choutouridis AEM:8997
 *    <cchoutou@ece.auth.gr>
 */

#include <vector>
#include <algorithm>
#include <cmath>
#include <cassert>
#include "distbitonic.hpp"

/*!
 * Returns the ascending or descending configuration of the node's sequence based on
 * the current node (MPI process) and the depth of the sorting network
 *
 * @param node      The current node (MPI process)
 * @param depth     The total depth of the sorting network (same for each step for a given network)
 *
 * @return          True if we need ascending configuration, false otherwise
 */
bool ascending(size_t node, size_t depth) noexcept {
    return !(node & (1 << depth));
}

/*!
 * Returns the node's partner for data exchange during the sorting network iterations
 *
 * @param node      The current node
 * @param step      The step of the sorting network
 * @return          The node id of the partner for data exchange
 */
size_t partner(size_t node, size_t step) noexcept {
    return (node ^ (1 << step));
}

/*!
 * Predicate to check if a node keeps the small numbers during the bitonic
 * sort network exchange.
 *
 * @param node      The node for which we check
 * @param partner   The partner of the data exchange
 * @param depth     The total depth of the sorting network (same for each step for a given network)
 * @return          True if the node should keep the small values, false otherwise
 */
bool keepsmall(size_t node, size_t partner, size_t depth) noexcept {
    assert(node != partner);
    return ascending(node, depth) == (node < partner);
}

void exchange(size_t node, size_t partner) {
    assert(node != partner);

}

void minmax(AllData_t& data, size_t node, size_t partner, bool keepsmall) {
    for (size_t i = 0; i < data[node].size(); ++i) {
        if (keepsmall && data[node][i] > data[partner][i])
            std::swap(data[node][i], data[partner][i]);
        if (!keepsmall && data[node][i] < data[partner][i])
            std::swap(data[node][i], data[partner][i]);
    }
}

void sort_network(AllData_t& data, size_t nodes, size_t depth) {
    for (size_t step = depth; step > 0;) {
        --step;
        for (size_t node = 0; node < nodes; ++node) {
            auto part = partner(node, step);
            auto ks = keepsmall(node, part, depth);
            if (node < part) {
                exchange(node, part);
                minmax(data, node, part, ks);
            }
        }
    }
}

void distbitonic(size_t P, AllData_t& data) {
    auto p = static_cast<uint32_t>(std::log2(P));

    for (size_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
        // Initially sort to create the half part of a bitonic
        if (ascending(node, 0))
            std::sort(data[node].begin(), data[node].end(), std::less<>());
        else
            std::sort(data[node].begin(), data[node].end(), std::greater<>());
    }

    // Run through sort network using elbow-sort
    for (size_t depth = 1; depth <= p; ++depth) {
        sort_network(data, P, depth);

        for (size_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
            // elbow-sort here
            if (ascending(node, depth))
                std::sort(data[node].begin(), data[node].end(), std::less<>());
            else
                std::sort(data[node].begin(), data[node].end(), std::greater<>());
        }

    }
}
